@@ -4,10 +4,13 @@ from .EconDensity import EconDensity
 from .ArmaJump import ArmaJump
 from .JumpDiffusionModel import JumpDiffusionModel
 from .SkewNormal import SkewNormal
+from .LinearGaussian import LinearGaussian
+from .LinearStudentT import LinearStudentT
+
 import inspect, sys
 import numpy as np
 
-def get_probabilistic_models_list():
-  clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
-  return np.asarray(clsmembers)
+# def get_probabilistic_models_list():
+#   clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
+#   return np.asarray(clsmembers)
 
