@@ -11,7 +11,7 @@ from tensorflow.python.training import moving_averages
 from collections import OrderedDict
 from collections import deque
 from itertools import chain
-from inspect import getargspec
+from inspect import getfullargspec
 from difflib import get_close_matches
 from warnings import warn
 
@@ -707,7 +707,10 @@ class SpatialExpectedSoftmaxLayer(Layer):
 
 
 class DropoutLayer(Layer):
-    def __init__(self, incoming, p=0.5, rescale=True, **kwargs):
+    def __init__(self, incoming, p, rescale=False, **kwargs):
+        """
+        :param p: probability of setting the output of a node to 0. Should be a tf placeholder
+        """
         super(DropoutLayer, self).__init__(incoming, **kwargs)
         self.p = p
         self.rescale = rescale
@@ -724,9 +727,7 @@ class DropoutLayer(Layer):
         if deterministic or self.p == 0:
             return input
         else:
-            # Using theano constant to prevent upcasting
-            # one = T.constant(1)
-
+            # I don't know what this is for. TensorFlow rescales automatically when using tf.nn.dropout
             retain_prob = 1. - self.p
             if self.rescale:
                 input /= retain_prob
@@ -738,10 +739,17 @@ class DropoutLayer(Layer):
         return input_shape
 
 class GaussianNoiseLayer(Layer):
-    def __init__(self, incoming, noise_std, noise_on_ph=None, **kwargs):
+    def __init__(self, incoming, noise_std, noise_on_ph=None, name='', **kwargs):
         super(GaussianNoiseLayer, self).__init__(incoming, **kwargs)
-        self.noise_std = noise_std
-        self.noise_on_ph = noise_on_ph # boolean placeholder that controlls whether noise shall be added or not (i.e. train vs. test time)
+
+        self.noise_std = self.add_param(
+            tf.Variable(noise_std, dtype=tf.float32, name='noise_std_' + name, trainable=False),
+            (1,),
+            name='noise_std_' + name,
+            trainable=False
+        )
+        # boolean placeholder that controlls whether noise shall be added or not (i.e. train vs. test time)
+        self.noise_on_ph = noise_on_ph
 
     def get_output_for(self, input, **kwargs):
         if self.noise_on_ph is not None:
@@ -1130,7 +1138,7 @@ class GRUStepLayer(MergeLayer):
         self._gru_layer = recurrent_layer
 
     def get_params(self, **tags):
-        return self._gru_layer.get_params(**tags)
+        return self._gru_layer._get_params(**tags)
 
     def get_output_shape_for(self, input_shapes):
         n_batch = input_shapes[0][0]
@@ -1562,7 +1570,7 @@ class LSTMStepLayer(MergeLayer):
         self._recurrent_layer = recurrent_layer
 
     def get_params(self, **tags):
-        return self._recurrent_layer.get_params(**tags)
+        return self._recurrent_layer._get_params(**tags)
 
     def get_output_shape_for(self, input_shapes):
         n_batch = input_shapes[0][0]
@@ -1858,7 +1866,7 @@ def get_output(layer_or_layers, inputs=None, **kwargs):
                                  % layer)
             all_outputs[layer] = layer.get_output_for(layer_inputs, **kwargs)
             try:
-                names, _, _, defaults = getargspec(layer.get_output_for)
+                names, _, _, defaults, _, _, _ = getfullargspec(layer.get_output_for)
             except TypeError:
                 # If introspection is not possible, skip it
                 pass
