@@ -9,6 +9,6 @@ def flatten_tensors(tensors):
 
 
 def unflatten_tensors(flattened, tensor_shapes):
-    tensor_sizes = list(map(np.prod, tensor_shapes))
+    tensor_sizes = list(map(int, map(np.prod, tensor_shapes)))
     indices = np.cumsum(tensor_sizes)[:-1]
     return [np.reshape(pair[0], pair[1]) for pair in zip(np.split(flattened, indices), tensor_shapes)]
\ No newline at end of file
