@@ -9,76 +9,8 @@ import numpy as np
 import scipy.stats as stats
 
 sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
-from cde.helpers import sample_center_points, mc_integration_cauchy, norm_along_axis_1
 from cde.density_estimator import MixtureDensityNetwork, KernelMixtureNetwork, \
-  ConditionalKernelDensityEstimation, LSConditionalDensityEstimation, NeighborKernelDensityEstimation
-
-class TestHelpers(unittest.TestCase):
-
-  """ sample center points """
-
-  def test_1_shape_center_point(self):
-    methods = ["all", "random", "k_means" , "agglomerative"]
-    for m in methods:
-      Y = np.random.uniform(size=(120,2))
-      centers = sample_center_points(Y, method=m, k=50)
-      self.assertEqual(centers.ndim, Y.ndim)
-      self.assertEqual(centers.shape[1], Y.shape[1])
-
-  def test_1_shape_center_point_k_means(self):
-    Y = np.asarray([1.0, 2.0])
-    centers = sample_center_points(Y, method="k_means", k=1)
-    self.assertAlmostEqual(Y.mean(), centers.mean())
-
-  def test_1_shape_center_point_agglomerative(self):
-    Y = np.random.uniform(size=[20,3])
-    centers = sample_center_points(Y, method="agglomerative", k=1)
-    self.assertAlmostEqual(Y.mean(), centers.mean())
-
-  def test_1_shape_center_point_keep_edges(self):
-    methods = ["random", "k_means", "agglomerative"]
-    for m in methods:
-      Y = np.random.uniform(size=(100,2))
-      centers = sample_center_points(Y, method=m, k=5, keep_edges=True)
-      self.assertEqual(centers.ndim, Y.ndim)
-      self.assertEqual(centers.shape[1], Y.shape[1])
-
-  """ norm along axis """
-
-  def test_2_norm_along_axis_1(self):
-    A = np.asarray([[1.0, 0.0], [1.0, 0.0]])
-    B = np.asarray([[0.0,0.0], [0.0,0.0]])
-    dist1 = norm_along_axis_1(A, B, squared=True)
-    dist2 = norm_along_axis_1(A, B, squared=False)
-    self.assertEqual(np.mean(dist1), 1.0)
-    self.assertEqual(np.mean(dist2), 1.0)
-
-  def test_2_norm_along_axis_2(self):
-    A = np.asarray([[1.0, 0.0]])
-    B = np.asarray([[0.0,0.0]])
-    dist1 = norm_along_axis_1(A, B, squared=True)
-    dist2 = norm_along_axis_1(A, B, squared=False)
-    self.assertEqual(np.mean(dist1), 1.0)
-    self.assertEqual(np.mean(dist2), 1.0)
-
-  def test_2_norm_along_axis_3(self):
-    A = np.random.uniform(size=[20, 3])
-    B = np.random.uniform(size=[10, 3])
-    dist = norm_along_axis_1(A, B, squared=True)
-    self.assertEqual(dist.shape, (20,10))
-
-  """ monte carlo integration """
-
-  def test_mc_integration_chauchy_1(self):
-    func = lambda y: np.expand_dims(stats.multivariate_normal.pdf(y, mean=[0, 0], cov=np.diag([2, 2])), axis=1)
-    integral = mc_integration_cauchy(func, ndim=2, n_samples=10 ** 7, batch_size=10**6)
-    self.assertAlmostEqual(1.0, integral[0], places=2)
-
-  def test_mc_integration_chauchy_2(self):
-    func = lambda y: y * np.tile(np.expand_dims(stats.multivariate_normal.pdf(y, mean=[1, 2], cov=np.diag([2, 2])), axis=1), (1,2))
-    integral = mc_integration_cauchy(func, ndim=2, n_samples=10 ** 7, batch_size=10**6)
-    self.assertAlmostEqual(1, integral[0], places=2)
-    self.assertAlmostEqual(2, integral[1], places=2)
+  ConditionalKernelDensityEstimation, LSConditionalDensityEstimation, NeighborKernelDensityEstimation, NormalizingFlowEstimator
 
 class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
 
@@ -89,7 +21,7 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
     Y = data[:, 1]
     return X, Y
 
-  def test_NKDE_with_2d_gaussian(self):
+  def test_NKDE_with_4d_gaussian(self):
     mu = 5
     std = 2.0
     X = np.random.normal(loc=mu, scale=std, size=(4000, 2))
@@ -99,10 +31,10 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
     model.fit(X, Y)
 
     y = np.random.uniform(low=[1.0, 1.0], high=[9.0, 9.0], size=(500, 2))
-    x = np.ones(shape=(500,2)) * mu
+    x = np.ones(shape=(500, 2)) * mu
 
     p_est = model.pdf(x, y)
-    p_true = norm.pdf(y, loc=mu, scale=std)
+    p_true = stats.multivariate_normal.pdf(y, mean=np.ones(2)*mu, cov=std**2)
 
     self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
 
@@ -128,19 +60,90 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
       p_true = norm.pdf(y, loc=-1, scale=1.5)
       self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
 
-  def test_LSCD_with_2d_gaussian(self):
-    X, Y = self.get_samples()
+  def test_NKDE_loo_log_likelihood(self):
+    mu = 1
+    std = 1
+    X = np.random.normal(loc=mu, scale=std, size=(500, 2))
+    Y = np.random.normal(loc=mu, scale=std, size=(500, 2))
+
+    model = NeighborKernelDensityEstimation(epsilon=0.3)
+    model.fit(X, Y)
+    bw = np.array([0.5])
+    epsilon = 0.3
+    ll1 = model.loo_likelihood(bandwidth=bw, epsilon=epsilon)
+
+    bw = np.array([0.05])
+    epsilon = 0.1
+    ll2 = model.loo_likelihood(bandwidth=bw, epsilon=epsilon)
+    self.assertGreater(ll1, ll2)
+
+  def test_NKDE_param_selection(self):
+    mu = 5
+    std = 2
+    X = np.random.normal(loc=mu, scale=std, size=(500, 2))
+    Y = np.random.normal(loc=mu, scale=std, size=(500, 2))
+
+    model1 = NeighborKernelDensityEstimation('NKDE', 2, 2, epsilon=0.1, bandwidth=0.3, param_selection=None)
+    model1.fit(X, Y)
+    self.assertAlmostEqual(model1.epsilon, 0.1)
+
+    model2 = NeighborKernelDensityEstimation('NKDE', 2, 2, epsilon=0.1, bandwidth=0.3, param_selection='normal_reference')
+    model2.fit(X, Y)
+    self.assertAlmostEqual(model1.epsilon, 0.1)
+    self.assertGreaterEqual(np.mean(model2.bandwidth - model1.bandwidth), 0.0)
+
+    model3 = NeighborKernelDensityEstimation('NKDE', 2, 2, epsilon=0.1, bandwidth=0.3, param_selection='cv_ml')
+    model3.fit(X, Y)
+    self.assertNotEqual(model3.epsilon, 0.1)
+
+    X_test, Y_test = np.random.normal(loc=mu, scale=std, size=(2000, 2)), np.random.normal(loc=mu, scale=std, size=(2000, 2))
+    score1 = model1.score(X_test, Y_test)
+    score2 = model2.score(X_test, Y_test)
+    score3 = model3.score(X_test, Y_test)
+    print(score1, score2, score3)
+
+    self.assertGreaterEqual(score2, score1)
+    self.assertGreaterEqual(score3, score2 - 0.1)
+
+  def test_LSCD_with_4d_gaussian(self):
+    mu = 5
+    std = 2.0
+    X = np.random.normal(loc=mu, scale=std, size=(4000, 2))
+    Y = np.random.normal(loc=mu, scale=std, size=(4000, 2))
 
     for method in ["all", "k_means"]:
-      model = LSConditionalDensityEstimation(center_sampling_method=method, n_centers=400, bandwidth=0.5)
+      model = LSConditionalDensityEstimation(center_sampling_method=method)
       model.fit(X, Y)
 
-      y = np.arange(-1, 5, 0.5)
-      x = np.asarray([2 for i in range(y.shape[0])])
+      y = np.random.uniform(low=[1.0, 1.0], high=[9.0, 9.0], size=(500, 2))
+      x = np.ones(shape=(500, 2)) * mu
+
       p_est = model.pdf(x, y)
-      p_true = norm.pdf(y, loc=2, scale=1)
+      p_true = stats.multivariate_normal.pdf(y, mean=np.ones(2)*mu, cov=std**2)
+
       self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
 
+  def test_LSCD_with_2d_gaussian(self):
+    X = np.random.uniform(-1, 1, size=4000)
+    Y = (2 + X) * np.random.normal(size=4000) + 2*X
+
+    model = LSConditionalDensityEstimation()
+    model.fit(X, Y)
+
+    y = np.linspace(-5, 5, num=100)
+    x = np.ones(100) * 0
+
+    p_est = model.pdf(x, y)
+    p_true = norm.pdf(y, loc=0, scale=2)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
+
+    y = np.linspace(-5, 5, num=100)
+    x = - np.ones(100) * 0.5
+
+    p_est = model.pdf(x, y)
+    p_true = norm.pdf(y, loc=-1, scale=1.5)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
+
   def test_KMN_with_2d_gaussian(self):
     mu = -2.0
     std = 2.0
@@ -192,7 +195,7 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
     import time
     t = time.time()
     model = KernelMixtureNetwork("kmn_sampling", 1, 1, center_sampling_method='k_means', n_centers=5,
-                                 n_training_epochs=500, data_normalization=False)
+                                 n_training_epochs=1000, data_normalization=True)
     print("time to build model:", time.time() - t)
     t = time.time()
 
@@ -240,6 +243,7 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
     p_true = norm.cdf(y, loc=mu, scale=std)
     self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
 
+
   def test_MDN_with_2d_gaussian2(self):
     mu = -5
     std = 2.5
@@ -274,6 +278,19 @@ class TestConditionalDensityEstimators_2d_gaussian(unittest.TestCase):
     p_true = norm.cdf(y, loc=2, scale=1)
     self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
 
+  def test_MDN_KMN_eval_set(self):
+    mu = 200
+    std = 23
+    X_train, Y_train = self.get_samples(mu=mu, std=std)
+    X_test, Y_test = self.get_samples(mu=mu, std=std)
+    X_test = X_test
+
+    model = MixtureDensityNetwork("mdn_eval_set", 1, 1, n_centers=10, data_normalization=True, n_training_epochs=100)
+    model.fit(X_train, Y_train, eval_set=(X_test, Y_test))
+
+    model = KernelMixtureNetwork("kmn_eval_set", 1, 1, n_centers=10, data_normalization=True, n_training_epochs=100)
+    model.fit(X_train, Y_train, eval_set=(X_test, Y_test))
+
 class TestSerializationDensityEstimators(unittest.TestCase):
 
   def get_samples(self, std=1.0):
@@ -319,9 +336,9 @@ class TestSerializationDensityEstimators(unittest.TestCase):
 
 class TestRegularization(unittest.TestCase):
 
-  def get_samples(self, std=1.0, mean=2):
+  def get_samples(self, std=1.0, mu=2, n_samples=2000):
     np.random.seed(22)
-    data = np.random.normal([mean, mean], std, size=(2000, 2))
+    data = np.random.normal([mu, mu], std, size=(n_samples, 2))
     X = data[:, 0]
     Y = data[:, 1]
     return X, Y
@@ -329,18 +346,19 @@ class TestRegularization(unittest.TestCase):
   def test1_KMN_with_2d_gaussian_noise_y(self):
     X, Y = self.get_samples(std=0.5)
 
-    model_no_noise = KernelMixtureNetwork("kmn_no_noise_y", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=None)
-    model_no_noise.fit(X, Y)
-    var_no_noise = model_no_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
+    with tf.Session():
+      model_no_noise = KernelMixtureNetwork("kmn_no_noise_y", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=None)
+      model_no_noise.fit(X, Y)
+      var_no_noise = model_no_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
 
-    model_noise = KernelMixtureNetwork("kmn_noise_y", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=1)
-    model_noise.fit(X, Y)
-    var_noise = model_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
+      model_noise = KernelMixtureNetwork("kmn_noise_y", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=1)
+      model_noise.fit(X, Y)
+      var_noise = model_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
 
-    print("Training w/o noise:", var_no_noise)
-    print("Training w/ noise:", var_noise)
+      print("Training w/o noise:", var_no_noise)
+      print("Training w/ noise:", var_noise)
 
-    self.assertGreaterEqual(var_noise - var_no_noise, 0.1)
+      self.assertGreaterEqual(var_noise - var_no_noise, 0.1)
 
   def test2_KMN_with_2d_gaussian_noise_x(self):
     np.random.seed(22)
@@ -351,34 +369,36 @@ class TestRegularization(unittest.TestCase):
     x_test_4 = np.ones(100) * 4
     y_test = np.linspace(1, 5, num=100)
 
-    model_no_noise = KernelMixtureNetwork("kmn_no_noise_x", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=None)
-    model_no_noise.fit(X, Y)
-    pdf_distance_no_noise = np.mean(np.abs(model_no_noise.pdf(x_test_2, y_test) - model_no_noise.pdf(x_test_4, y_test)))
+    with tf.Session():
+      model_no_noise = KernelMixtureNetwork("kmn_no_noise_x", 1, 1, n_centers=5, x_noise_std=None, y_noise_std=None)
+      model_no_noise.fit(X, Y)
+      pdf_distance_no_noise = np.mean(np.abs(model_no_noise.pdf(x_test_2, y_test) - model_no_noise.pdf(x_test_4, y_test)))
 
-    model_noise = KernelMixtureNetwork("kmn_noise_x", 1, 1, n_centers=5, x_noise_std=2, y_noise_std=None)
-    model_noise.fit(X, Y)
-    pdf_distance_noise = np.mean(np.abs(model_noise.pdf(x_test_2, y_test) - model_noise.pdf(x_test_4, y_test)))
+      model_noise = KernelMixtureNetwork("kmn_noise_x", 1, 1, n_centers=5, x_noise_std=2, y_noise_std=None)
+      model_noise.fit(X, Y)
+      pdf_distance_noise = np.mean(np.abs(model_noise.pdf(x_test_2, y_test) - model_noise.pdf(x_test_4, y_test)))
 
-    print("Training w/o noise - pdf distance:", pdf_distance_no_noise)
-    print("Training w/ noise - pdf distance", pdf_distance_noise)
+      print("Training w/o noise - pdf distance:", pdf_distance_no_noise)
+      print("Training w/ noise - pdf distance", pdf_distance_noise)
 
-    self.assertGreaterEqual(pdf_distance_no_noise / pdf_distance_noise, 2.0)
+      self.assertGreaterEqual(pdf_distance_no_noise / pdf_distance_noise, 2.0)
 
   def test3_MDN_with_2d_gaussian_noise_y(self):
     X, Y = self.get_samples(std=0.5)
 
-    model_no_noise = MixtureDensityNetwork("mdn_no_noise_y", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=None)
-    model_no_noise.fit(X, Y)
-    var_no_noise = model_no_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
+    with tf.Session():
+      model_no_noise = MixtureDensityNetwork("mdn_no_noise_y", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=None)
+      model_no_noise.fit(X, Y)
+      var_no_noise = model_no_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
 
-    model_noise = MixtureDensityNetwork("mdn_noise_y", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=1)
-    model_noise.fit(X, Y)
-    var_noise = model_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
+      model_noise = MixtureDensityNetwork("mdn_noise_y", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=1)
+      model_noise.fit(X, Y)
+      var_noise = model_noise.covariance(x_cond=np.array([[2]]))[0][0][0]
 
-    print("Training w/o noise:", var_no_noise)
-    print("Training w/ noise:", var_noise)
+      print("Training w/o noise:", var_no_noise)
+      print("Training w/ noise:", var_noise)
 
-    self.assertGreaterEqual(var_noise - var_no_noise, 0.1)
+      self.assertGreaterEqual(var_noise - var_no_noise, 0.1)
 
   def test4_MDN_with_2d_gaussian_noise_x(self):
     np.random.seed(22)
@@ -389,71 +409,22 @@ class TestRegularization(unittest.TestCase):
     x_test_4 = np.ones(100) * 4
     y_test = np.linspace(1,5,num=100)
 
-    model_no_noise = MixtureDensityNetwork("mdn_no_noise_x", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=None)
-    model_no_noise.fit(X, Y)
-    pdf_distance_no_noise = np.mean(np.abs(model_no_noise.pdf(x_test_2, y_test) - model_no_noise.pdf(x_test_4, y_test)))
+    with tf.Session():
+      model_no_noise = MixtureDensityNetwork("mdn_no_noise_x", 1, 1, n_centers=1, x_noise_std=None, y_noise_std=None)
+      model_no_noise.fit(X, Y)
+      pdf_distance_no_noise = np.mean(np.abs(model_no_noise.pdf(x_test_2, y_test) - model_no_noise.pdf(x_test_4, y_test)))
 
-    model_noise = MixtureDensityNetwork("mdn_noise_x", 1, 1, n_centers=1, x_noise_std=2, y_noise_std=None)
-    model_noise.fit(X, Y)
-    pdf_distance_noise = np.mean(np.abs(model_noise.pdf(x_test_2, y_test) - model_noise.pdf(x_test_4, y_test)))
+      model_noise = MixtureDensityNetwork("mdn_noise_x", 1, 1, n_centers=1, x_noise_std=2, y_noise_std=None)
+      model_noise.fit(X, Y)
+      pdf_distance_noise = np.mean(np.abs(model_noise.pdf(x_test_2, y_test) - model_noise.pdf(x_test_4, y_test)))
 
-    print("Training w/o noise - pdf distance:", pdf_distance_no_noise)
-    print("Training w/ noise - pdf distance", pdf_distance_noise)
+      print("Training w/o noise - pdf distance:", pdf_distance_no_noise)
+      print("Training w/ noise - pdf distance", pdf_distance_noise)
 
-    self.assertGreaterEqual(pdf_distance_no_noise/pdf_distance_noise, 2.0)
-
-  def test5_MDN_entropy_regularization(self):
-    X1, Y1 = self.get_samples(std=1, mean=2)
-    X2, Y2 = self.get_samples(std=1, mean=-2)
-
-    # DATA for GMM with two modes
-    X = np.expand_dims(np.concatenate([X1, X2], axis=0), axis=1)
-    Y = np.expand_dims(np.concatenate([Y1, Y2], axis=0), axis=1)
-
-    with tf.Session() as sess:
-      model_no_reg = MixtureDensityNetwork("mdn_no_entropy_reg", 1, 1, n_centers=2, x_noise_std=None, y_noise_std=None,
-                                          entropy_reg_coef=0.0)
-      model_no_reg.fit(X, Y)
-
-      entropy1 = np.mean(
-        sess.run(model_no_reg.softmax_entropy, feed_dict={model_no_reg.X_ph: X, model_no_reg.Y_ph: Y}))
-
-      model_reg = MixtureDensityNetwork("mdn_entropy_reg", 1, 1, n_centers=2, x_noise_std=None, y_noise_std=None,
-                                        entropy_reg_coef=10.0)
-      model_reg.fit(X, Y)
-      entropy2 = np.mean(sess.run(model_reg.softmax_entropy, feed_dict={model_reg.X_ph: X, model_reg.Y_ph: Y}))
-
-      print(entropy1)
-      print(entropy2)
-      self.assertGreaterEqual(entropy1 / entropy2, 10)
-
-  def test6_KMN_entropy_regularization(self):
-    X1, Y1 = self.get_samples(std=1, mean=2)
-    X2, Y2 = self.get_samples(std=1, mean=-2)
-
-    # DATA for GMM with two modes
-    X = np.expand_dims(np.concatenate([X1, X2], axis=0), axis=1)
-    Y = np.expand_dims(np.concatenate([Y1, Y2], axis=0), axis=1)
-
-    with tf.Session() as sess:
-      model_no_reg = KernelMixtureNetwork("kmn_no_entropy_reg", 1, 1, n_centers=2, x_noise_std=None, y_noise_std=None,
-                                          entropy_reg_coef=0.0)
-      model_no_reg.fit(X, Y)
-
-      entropy1 = np.mean(
-        sess.run(model_no_reg.softmax_entropy, feed_dict={model_no_reg.X_ph: X, model_no_reg.Y_ph: Y}))
-
-      model_reg = KernelMixtureNetwork("kmn_entropy_reg", 1, 1, n_centers=2, x_noise_std=None, y_noise_std=None,
-                                        entropy_reg_coef=10.0)
-      model_reg.fit(X, Y)
-      entropy2 = np.mean(sess.run(model_reg.softmax_entropy, feed_dict={model_reg.X_ph: X, model_reg.Y_ph: Y}))
-
-      print(entropy1)
-      print(entropy2)
-      self.assertGreaterEqual(entropy1 / entropy2, 10)
+      self.assertGreaterEqual(pdf_distance_no_noise/pdf_distance_noise, 2.0)
 
   def test7_data_normalization(self):
-    X, Y = self.get_samples(std=2, mean=20)
+    X, Y = self.get_samples(std=2, mu=20)
     with tf.Session() as sess:
       model = KernelMixtureNetwork("kmn_data_normalization", 1, 1, n_centers=2, x_noise_std=None, y_noise_std=None,
                                           data_normalization=True, n_training_epochs=100)
@@ -518,70 +489,244 @@ class TestRegularization(unittest.TestCase):
       self.assertGreaterEqual(cond_cov[1][1], std**2 * 0.7)
       self.assertLessEqual(cond_cov[1][1], std**2 * 1.3)
 
+  def test_MDN_adaptive_noise(self):
+    adaptive_noise_fn = lambda n, d: 0.0 if n < 1000 else 5.0
 
-#
-# class TestConditionalDensityEstimators_fit_by_crossval(unittest.TestCase):
-#   def get_samples(self):
-#     np.random.seed(22)
-#     data = np.concatenate([np.random.normal([i, -i], 1, size=(500, 2)) for i in range(-20, 20, 4)], axis=0)
-#     X = data[:, 0]
-#     Y = data[:, 1]
-#     return X, Y
-#
-#   def test_1_KMN_with_2d_gaussian_fit_by_crossval(self):
-#     X, Y = self.get_samples()
-#
-#     param_grid = {
-#       "n_centers": [3, 10],
-#       "center_sampling_method": ["k_means"],
-#       "keep_edges": [True]
-#     }
-#
-#     model = KernelMixtureNetwork(center_sampling_method="k_means", n_centers=20)
-#     model.fit_by_cv(X, Y, param_grid=param_grid)
-#
-#     y = np.arange(-1, 5, 0.5)
-#     x = np.asarray([2 for i in range(y.shape[0])])
-#     p_est = model.pdf(x, y)
-#     p_true = norm.pdf(y, loc=2, scale=1)
-#     self.assertEqual(model.get_params()["n_centers"], 10)
-#     self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.2)
-#
-#
-#   def test_2_MDN_with_2d_gaussian_fit_by_crossval(self):
-#     X, Y = self.get_samples()
-#
-#     param_grid = {
-#       "n_centers": [2, 10, 50]
-#     }
-#
-#     model = MixtureDensityNetwork()
-#     model.fit_by_cv(X, Y, param_grid=param_grid)
-#
-#     y = np.arange(-1, 5, 0.5)
-#     x = np.asarray([2 for i in range(y.shape[0])])
-#     p_est = model.pdf(x, y)
-#     p_true = norm.pdf(y, loc=2, scale=1)
-#     self.assertEqual(model.get_params()["n_centers"], 10)
-#     self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.2)
+    X, Y = self.get_samples(mu=0, std=1, n_samples=999)
+    est = MixtureDensityNetwork("mdn_adaptive_noise_999", 1, 1, n_centers=1, y_noise_std=0.0, x_noise_std=0.0,
+                                hidden_sizes=(8, 8),
+                                adaptive_noise_fn=adaptive_noise_fn, n_training_epochs=500)
+    est.fit(X, Y)
+    std_999 = est.std_(x_cond=np.array([[0.0]]))[0]
+
+    X, Y = self.get_samples(mu=0, std=1, n_samples=1002)
+    est = MixtureDensityNetwork("mdn_adaptive_noise_1002", 1, 1, n_centers=1, y_noise_std=0.0, x_noise_std=0.0,
+                                hidden_sizes=(8, 8),
+                                adaptive_noise_fn=adaptive_noise_fn, n_training_epochs=500)
+    est.fit(X, Y)
+    std_1002 = est.std_(x_cond=np.array([[0.0]]))[0]
+
+    self.assertLess(std_999, std_1002)
+    self.assertGreater(std_1002, 2)
+
+  def test_NF_adaptive_noise(self):
+    adaptive_noise_fn = lambda n, d: 0.0 if n < 1000 else 5.0
+
+    X, Y = self.get_samples(mu=0, std=1, n_samples=999)
+    est = NormalizingFlowEstimator("nf_999", 1, 1, y_noise_std=0.0, n_flows=2, hidden_sizes=(8,8),
+                                   x_noise_std=0.0, adaptive_noise_fn=adaptive_noise_fn,
+                                   n_training_epochs=500)
+    est.fit(X, Y)
+    std_999 = est.std_(x_cond=np.array([[0.0]]))[0]
+
+    X, Y = self.get_samples(mu=0, std=1, n_samples=1002)
+    est = NormalizingFlowEstimator("nf_1002", 1, 1, y_noise_std=0.0, n_flows=2,  hidden_sizes=(8,8),
+                                   x_noise_std=0.0, adaptive_noise_fn=adaptive_noise_fn,
+                                   n_training_epochs=500)
+    est.fit(X, Y)
+    std_1002 = est.std_(x_cond=np.array([[0.0]]))[0]
+
+    self.assertLess(std_999, std_1002)
+    self.assertGreater(std_1002, 2)
+
+  def test_MDN_weight_decay(self):
+    mu = 5
+    std = 5
+    X, Y = self.get_samples(mu=mu, std=std)
+
+    no_decay = MixtureDensityNetwork("mdn_no_weight_decay", 1, 1, hidden_sizes=(32, 32), n_centers=10,
+                                     n_training_epochs=2000, weight_decay=0.0, weight_normalization=False)
+    decay = MixtureDensityNetwork("mdn_weight_decay", 1, 1, n_centers=10,  hidden_sizes=(32, 32),
+                                  n_training_epochs=2000, weight_decay=1e-3, weight_normalization=False)
+    no_decay.fit(X, Y)
+    decay.fit(X, Y)
+
+    y = np.arange(mu - 3 * std, mu + 3 * std, 6 * std / 20)
+    x = np.asarray([mu for i in range(y.shape[0])])
+    p_true = norm.pdf(y, loc=mu, scale=std)
+    l1_err_no_dec = np.mean(np.abs(no_decay.pdf(x, y) - p_true))
+    l1_err_dec = np.mean(np.abs(decay.pdf(x, y) - p_true))
+
+    self.assertLessEqual(l1_err_dec, 0.1)
+    self.assertLessEqual(l1_err_dec, l1_err_no_dec)
+
+  def test_KMN_l2_regularization(self):
+    mu = 5
+    std = 5
+    X, Y = self.get_samples(mu=mu, std=std, n_samples=500)
+
+    kmn_no_reg = KernelMixtureNetwork("kmn_no_reg", 1, 1, n_centers=10,
+                                     n_training_epochs=200, l2_reg=0.0, weight_normalization=False)
+    kmn_reg_l2 = KernelMixtureNetwork("kmn_reg_l2", 1, 1, n_centers=10, hidden_sizes=(16, 16),
+                                  n_training_epochs=200, l2_reg=1.0, weight_normalization=False)
+    kmn_no_reg.fit(X, Y)
+    kmn_reg_l2.fit(X, Y)
+
+    y = np.arange(mu - 3 * std, mu + 3 * std, 6 * std / 20)
+    x = np.asarray([mu for i in range(y.shape[0])])
+    p_true = norm.pdf(y, loc=mu, scale=std)
+    err_no_reg = np.mean(np.abs(kmn_no_reg.pdf(x, y) - p_true))
+    err_reg_l2 = np.mean(np.abs(kmn_reg_l2.pdf(x, y) - p_true))
+
+    self.assertLessEqual(err_reg_l2, err_no_reg)
+
+  def test_NF_l1_regularization(self):
+    mu = 5
+    std = 5
+    X, Y = self.get_samples(mu=mu, std=std, n_samples=500)
+
+    nf_no_reg = NormalizingFlowEstimator("nf_no_reg", 1, 1, hidden_sizes=(16, 16), n_flows=10,
+                                         weight_normalization=False, l1_reg=0.0, l2_reg=0.0, n_training_epochs=500)
+    nf_reg_l1 = NormalizingFlowEstimator("nf_reg_l1", 1, 1, hidden_sizes=(16, 16), n_flows=10,
+                                         weight_normalization=False, l1_reg=10.0, l2_reg=0.0, n_training_epochs=500)
+    nf_no_reg.fit(X, Y)
+    nf_reg_l1.fit(X, Y)
+
+    y = np.arange(mu - 3 * std, mu + 3 * std, 6 * std / 20)
+    x = np.asarray([mu for i in range(y.shape[0])])
+    p_true = norm.pdf(y, loc=mu, scale=std)
+    err_no_reg = np.mean(np.abs(nf_no_reg.pdf(x, y) - p_true))
+    err_reg_l1 = np.mean(np.abs(nf_reg_l1.pdf(x, y) - p_true))
+
+    print(err_no_reg, err_reg_l1)
+
+    self.assertLessEqual(err_reg_l1, err_no_reg)
+
+  def test_MDN_dropout(self):
+    mu = -8
+    std = 2.5
+    X, Y = self.get_samples(mu=mu, std=std)
+
+    dropout_model = MixtureDensityNetwork("mdn_dropout_reasonable", 1, 1, n_centers=5, weight_normalization=True,
+                                          dropout=0.5, n_training_epochs=400)
+    dropout_model.fit(X, Y)
+
+    y = np.arange(mu - 3 * std, mu + 3 * std, 6 * std / 20)
+    x = np.asarray([mu for i in range(y.shape[0])])
+    p_est = dropout_model.pdf(x, y)
+    p_true = norm.pdf(y, loc=mu, scale=std)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
+
+    p_est = dropout_model.cdf(x, y)
+    p_true = norm.cdf(y, loc=mu, scale=std)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.1)
+
+class TestLogProbability(unittest.TestCase):
+
+  def test_KMN_log_pdf(self):
+    X, Y = np.random.normal(size=(1000, 3)), np.random.normal(size=(1000, 2))
+
+    for data_norm in [True, False]:
+      with tf.Session() as sess:
+        model = KernelMixtureNetwork("kmn_logprob"+str(data_norm), 3, 2, n_centers=5,
+                                     hidden_sizes=(8, 8), init_scales=np.array([0.5]), n_training_epochs=10, data_normalization=data_norm)
+        model.fit(X, Y)
+
+        x, y = np.random.normal(size=(1000, 3)), np.random.normal(size=(1000, 2))
+        prob = model.pdf(x,y)
+        log_prob = model.log_pdf(x,y)
+        self.assertLessEqual(np.mean(np.abs(prob - np.exp(log_prob))), 0.001)
+
+  def test_MDN_log_pdf(self):
+    X, Y = np.random.normal(size=(1000, 3)), np.random.normal(size=(1000, 2))
+
+    for data_norm in [True, False]:
+      with tf.Session() as sess:
+        model = MixtureDensityNetwork("mdn_logprob"+str(data_norm), 3, 2, n_centers=1,
+                                     hidden_sizes=(8, 8), n_training_epochs=10, data_normalization=data_norm)
+        model.fit(X, Y)
+
+        x, y = np.random.normal(size=(1000, 3)), np.random.normal(size=(1000, 2))
+        prob = model.pdf(x,y)
+        log_prob = model.log_pdf(x,y)
+        self.assertLessEqual(np.mean(np.abs(prob - np.exp(log_prob))), 0.001)
+
+  def test_CKDE_log_pdf(self):
+    X, Y = np.random.normal(size=(500, 2)), np.random.normal(size=(500, 2))
+
+    model = ConditionalKernelDensityEstimation(bandwidth='normal_reference')
+    model.fit(X, Y)
+
+    x, y = np.random.normal(size=(100, 2)), np.random.normal(size=(100, 2))
+    prob = model.pdf(x, y)
+    log_prob = model.log_pdf(x, y)
+    self.assertLessEqual(np.mean(np.abs(prob - np.exp(log_prob))), 0.001)
+
+  def test_NKDE_log_pdf(self):
+    X, Y = np.random.normal(size=(500, 2)), np.random.normal(size=(500, 2))
+
+    model = NeighborKernelDensityEstimation()
+    model.fit(X, Y)
+
+    x, y = np.random.normal(size=(100, 2)), np.random.normal(size=(100, 2))
+    prob = model.pdf(x, y)
+    log_prob = model.log_pdf(x, y)
+    self.assertLessEqual(np.mean(np.abs(prob - np.exp(log_prob))), 0.001)
+
+  def test_LSCDE_log_pdf(self):
+    X, Y = np.random.normal(size=(500, 2)), np.random.normal(size=(500, 2))
+
+    model = LSConditionalDensityEstimation()
+    model.fit(X, Y)
+
+    x, y = np.random.normal(size=(100, 2)), np.random.normal(size=(100, 2))
+    prob = model.pdf(x, y)
+    log_prob = model.log_pdf(x, y)
+    self.assertLessEqual(np.mean(np.abs(prob - np.exp(log_prob))), 0.001)
+
+class TestConditionalDensityEstimators_fit_by_crossval(unittest.TestCase):
+  def get_samples(self):
+    np.random.seed(22)
+    data = np.concatenate([np.random.normal([i, -i], 1, size=(500, 2)) for i in range(-20, 20, 4)], axis=0)
+    X = data[:, 0]
+    Y = data[:, 1]
+    return X, Y
+
+  def test_1_KMN_with_2d_gaussian_fit_by_crossval(self):
+    X, Y = self.get_samples()
+
+    param_grid = {
+      "n_centers": [3, 10],
+      "center_sampling_method": ["k_means"],
+      "keep_edges": [True]
+    }
+
+    model = KernelMixtureNetwork(center_sampling_method="k_means", n_centers=20)
+    model.fit_by_cv(X, Y, param_grid=param_grid)
+
+    y = np.arange(-1, 5, 0.5)
+    x = np.asarray([2 for i in range(y.shape[0])])
+    p_est = model.pdf(x, y)
+    p_true = norm.pdf(y, loc=2, scale=1)
+    self.assertEqual(model.get_params()["n_centers"], 10)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.2)
 
 
-def suite():
-  suite = unittest.TestSuite()
-  suite.addTest(TestConditionalDensityEstimators_2d_gaussian())
-  #suite.addTest(TestConditionalDensityEstimators_fit_by_crossval())
-  suite.addTest(TestHelpers())
-  return suite
+  def test_2_MDN_with_2d_gaussian_fit_by_crossval(self):
+    X, Y = self.get_samples()
+
+    param_grid = {
+      "n_centers": [2, 10, 50]
+    }
+
+    model = MixtureDensityNetwork()
+    model.fit_by_cv(X, Y, param_grid=param_grid)
+
+    y = np.arange(-1, 5, 0.5)
+    x = np.asarray([2 for i in range(y.shape[0])])
+    p_est = model.pdf(x, y)
+    p_true = norm.pdf(y, loc=2, scale=1)
+    self.assertEqual(model.get_params()["n_centers"], 10)
+    self.assertLessEqual(np.mean(np.abs(p_true - p_est)), 0.2)
 
 if __name__ == '__main__':
   warnings.filterwarnings("ignore")
 
   testmodules = [
-   'unittests_estimators.TestHelpers',
-   'unittests_estimators.TestRiskMeasures',
    'unittests_estimators.TestConditionalDensityEstimators_2d_gaussian',
    'unittests_estimators.TestRegularization',
-   #'unittests_estimators.TestConditionalDensityEstimators_fit_by_crossval'
+   'unittests_estimators.TestSerializationDensityEstimators',
+   'unittests_estimators.TestLogProbability'
    ]
   suite = unittest.TestSuite()
   for t in testmodules:
